#!/usr/bin/env python
from pyftpdlib import ftpserver
from aksyfs import common
from aksy.config import create_option_parser
from aksy.device import Devices
from aksy import fileutils

import fnmatch
import os, time, errno

import logging
LOG = logging.getLogger("aksy.aksyfs.ftpd")

class AksyFtpFS(common.AksyFS, ftpserver.AbstractedFS):
    def __call__(self):
        return AksyFtpFS(self.sampler)

    def __init__(self, sampler):
        common.AksyFS.__init__(self, sampler)
        self.root = ''
        self.cwd = '/'
        self.rnfr = None

    def translate(self, path):
        """Translate a 'virtual' FTP path into equivalent filesystem path. Take
        an absolute or relative path as input and return a full absolute file
        path.
        
        """
        if LOG.isEnabledFor(logging.DEBUG): 
            LOG.debug('translate %s', path)

        if not os.path.isabs(path):
            path = os.path.join(self.cwd, path)
        # For AksyFS only virtual paths exist
        return os.path.normpath(path).replace('\\', '/')

    def open(self, filename, mode):
        """Open a file returning its handler."""
        if LOG.isEnabledFor(logging.DEBUG): 
            LOG.debug('open(%s, %s)', filename, mode)
        if self.isdir(filename):
            raise IOError(errno.EINVAL, filename)
        if mode.find('r') != -1:
            return self.open_for_read(filename, mode)
        elif mode.find('w') != -1:
            return self.open_for_write(filename, mode)
        else:
            common.raiseException(errno.EINVAL)

    def exists(self, abspath):
        """Return True if the path exists."""
        try:
            self.getattr(abspath)
            return True
        except OSError:
            return False
        
    def isfile(self, abspath):
        """Return True if path is a file."""
        return not fileutils.is_dirpath(abspath)

    def isdir(self, abspath):
        """Return True if path is a directory."""
        return fileutils.is_dirpath(abspath)

    def chdir(self, abspath):
        """Change the current directory."""
        if LOG.isEnabledFor(logging.DEBUG): 
            LOG.debug("chdir ", abspath)
        i = 1
        while i < len(abspath):
            i = abspath.find('/', i)
            if i == -1: i = len(abspath)
            self.getattr(abspath[:i])
            i += 1
            
        self.cwd = abspath

    def remove(self, abspath):
        """Remove the specified file."""
        self.unlink(abspath)
    
    def getsize(self, abspath):
        """Return the size of the specified file in bytes."""
        return common.get_file_size(abspath)

    def getmtime(self, abspath):
        """Return the last modified time as a number of seconds since the
        epoch."""
        return common.START_TIME
           
    def glob1(self, dirname, pattern):
        """Return a list of files matching a dirname pattern non-recursively.
        Unlike glob.glob1 raises an exception if os.listdir() fails.
        """
        names = self.listdir(dirname)
        if pattern[0] != '.':
            names = filter(lambda x: x[0] != '.', names)
        return fnmatch.filter(names, pattern)

    # --- utility methods
    
    # Note that these are resource-intensive blocking operations so you may want
    # to override and move them into another process/thread in some way.

    def get_nlst_dir(self, abspath):
        """Return a directory listing in a form suitable for NLST command."""
        listing = '\r\n'.join(self.listdir(abspath))
        if listing:
            return listing + '\r\n'
        return ''

    def get_list_dir(self, abspath):
        """Return a directory listing in a form suitable for LIST command."""
        # if path is a file we return information about it
        if os.path.isfile(abspath):
            basedir, filename = os.path.split(abspath)
            listing = [filename]
        else:
            basedir = abspath
            listing = self.listdir(abspath)
        return self.format_list(basedir, listing)

    def format_list(self, basedir, listing):
        """Return a directory listing emulating "/bin/ls -lgA" UNIX command
        output.

        <basedir> is the absolute dirname, <listing> is a list of files
        contained in that directory.

        For portability reasons permissions, hard links numbers, owners and
        groups listed are static and unreliable but it shouldn't represent a
        problem for most ftp clients around.
        If you want reliable values on unix systems override this method and
        use other attributes provided by os.stat().
        This is how output appears to client:

        -rwxrwxrwx   1 owner    group         7045120 Sep 02  3:47 music.mp3
        drwxrwxrwx   1 owner    group               0 Aug 31 18:50 e-books
        -rwxrwxrwx   1 owner    group             380 Sep 02  3:40 module.py
        """
        result = []
        for basename in listing:
            f = os.path.join(basedir, basename)
            stat = self.getattr(f)

            # stat.st_mtime could fail (-1) if file's last modification time is
            # too old, in that case we return local time as last modification time.
            try:
                mtime = time.strftime("%b %d %H:%M", time.localtime(stat.st_mtime))
            except ValueError:
                mtime = time.strftime("%b %d %H:%M")

            if fileutils.is_dirpath(f):
                result.append("drwxrwxrwx   1 owner    group %15s %s %s\r\n" %(
                    '0', # no size
                    mtime, 
                    basename))
            else:
                result.append("-rw-rw-rw-   1 owner    group %15s %s %s\r\n" %(
                    stat.st_size, 
                    mtime, 
                    basename))
        return ''.join(result)

def main():
    parser = create_option_parser(usage='%prog [options]')
    parser.add_option("--ftp_host", nargs=1, dest="ftp_host",
      help="Server host or IP address to bind on. Only change if running in an isolated network!", default='localhost')
    parser.add_option("--ftp_port", nargs=1, dest="ftp_port",
      help="Server port to bind on.", default=2121)

    options = parser.parse_args()[0]

    authorizer = ftpserver.DummyAuthorizer()
    authorizer.add_anonymous('/', perm=('r', 'w'))

    address = (options.ftp_host, options.ftp_port)
    if options.ftp_host != 'localhost':
        import warnings
        warnings.warn('Running aksyftpd on non-local address %s' % options.ftp_host, RuntimeWarning)
    
    sampler = Devices.get_instance(options.sampler_type, options.connector)

    try:    
        ftp_handler = ftpserver.FTPHandler
        ftp_handler.authorizer = authorizer
        ftp_handler.banner = "aksyftpd (pyftpd version %s) ready." % ftpserver.__ver__
        ftp_handler.abstracted_fs = AksyFtpFS(sampler)

        ftpd = ftpserver.FTPServer(address, ftp_handler)
        ftpd.max_cons = 256
        ftpd.max_cons_per_ip = ftpd.max_cons 
        ftpd.serve_forever()
    finally:
        sampler.close()

if __name__ == "__main__":
    main()



"""Aksy setup module
"""

from setuptools import setup
from distutils.core import Extension
from distutils.dist import Distribution
from distutils.command.build_ext import build_ext
import platform, os.path, sys

version = "0.4" 

# macros= [("_DEBUG", 0), ("AKSY_DEBUG", "1")]
macros= [("AKSY_DEBUG", 0)]

def install_requires():
    deps = ["pyftpdlib == 0.3"]
    if not platform.system() == "Windows":
	deps.append("fuse-python >= 0.2pre3")
    return deps

def customize_for_platform(ext, compiler_type):
    ext.libraries = ["usb"]

    # Windows
    if platform.system() == "Windows":
        libusb_base_dir = "C:\Program Files\LibUSB-Win32"

    if compiler_type == "msvc":
        ext.libraries = ["libusb"]
        ext.extra_compile_args = ["/O2"]
        ext.library_dirs = [os.path.join(libusb_base_dir, "lib", "msvc")]

    if compiler_type == "mingw32":
        ext.libraries.append("mingw32")
        ext.library_dirs =[os.path.join(libusb_base_dir, "lib", "gcc")]
    
    # Unix flavours 
    if compiler_type == "unix":
        libusb_base_dir = "/usr/local/libusb-1.0"
        
    if platform.system() == "Darwin":
	libusb_base_dir = "/usr/local/Cellar/libusb-compat/0.1.5_1/"
        # ext.extra_link_args = ["-framework CoreFoundation IOKit"]

    ext.library_dirs = [os.path.join(libusb_base_dir, "lib")]
    ext.include_dirs = [os.path.join(libusb_base_dir, "include")]

class build_akyx(build_ext):
    def build_extension(self, ext):
        customize_for_platform(ext, self.compiler.compiler_type)
        build_ext.build_extension(self, ext)
        
aksy_packages = [
          "aksy", "aksyfs", "aksyosc", "aksy.devices", 
          "aksy.console",
          "aksy.devices.akai", 
          "aksy.devices.akai.mock_z48", 
          "aksy.devices.akai.z48", 
          "aksy.devices.akai.s56k"]

aksui_packages = ["aksui", "aksui.UI", "aksui.ak", "aksui.utils"]

all_packages = []
all_packages.extend(aksy_packages)
all_packages.extend(aksui_packages)

base_url = "https://github.com/watzo/aksy"

classifiers = """\
Development Status :: 3 - Alpha
Intended Audience :: Developers
Intended Audience :: End Users/Desktop
License :: OSI Approved :: GNU General Public License (GPL)
Programming Language :: Python
Programming Language :: C
Topic :: Multimedia :: Sound/Audio
Topic :: Software Development :: Libraries :: Python Modules
Topic :: System :: Filesystems
Topic :: System :: Hardware
Operating System :: Microsoft :: Windows
Operating System :: MacOS :: MacOS X
Operating System :: POSIX :: Linux
"""

setup(
      name = "aksy", 
      dependency_links = [base_url + "/aksy/dependencies/"],
      version = version,
      author = "Walco van Loon", 
      author_email = "walco+aksy@pitchdark.org", 
      description = "Control S5000/S6000, Z4/Z8 and MPC4000 Akai sampler models with System Exclusive over USB",
      license = "GPL",
      classifiers = filter(None, classifiers.split("\n")),
      package_dir = {"": "src"}, 
      packages = all_packages, 
      package_data = {"aksui": ["ak.py.glade"]},
      url = base_url + "/projects/aksy", 
      platforms = [ "any" ],
      download_url = base_url + "/downloads",
      install_requires = install_requires(),
      scripts = [],
      entry_points = {
        'console_scripts': [
            'aksy-get = aksy.console.get:main',
            'aksy-put = aksy.console.put:main',
            'aksy-fs = aksyfs.aksyfuse:main [FUSE-PYTHON]',
            'aksy-ftpd = aksyfs.ftpd:main [PYFTPDLIB]',
            'aksy-ui = aksui.main:main [PYGTK]',
        ],
        'gui_scripts': [
            'aksy-ui = aksui.main:main',
        ]
      },
      ext_modules = [
          Extension("aksyx",
              sources = [ "src/aksyx/aksyx.c", "src/aksyx/aksyxusb.c",],
              define_macros= macros
          ),
      ],
      cmdclass = {
         "build_ext": build_akyx, 
      },
      extras_require = {
        'FUSE-PYTHON':  ["fuse-python >= 0.2pre3"],
        'PYFTPDLIB' : ["pyftpdlib == 0.3"],
        'PYGTK' : ["pygtk"]
      },
      test_suite = "tests"
)



import samplerobject

class Part(samplerobject.SamplerObject):
    def __init__(self, s, m, index):
        samplerobject.SamplerObject.__init__(self, s, m, "multitools", index)
        self.attrs = ["multi_part_name", "part_midi_channel", "part_mute", "part_solo", "part_level", "part_output", "part_pan", "part_fx_channel", "part_fx_send_level", "part_tune", "part_low_note", "part_high_note", "part_priority", "part_prog_no", "part_group_id"]
        self.abbr = { "part_pan" : "pan" }
        self.need_index_for_set = True
        self.m = m
        
        self.precache()
        
    def get_handle(self):
        return self.m.get_handle()



from optparse import OptionParser

def create_option_parser(): 
    usage = "%prog"
    parser = OptionParser(usage=usage)
    parser.add_option("-p", type="int", dest="port", 
          help="Port number", default=6575)
    parser.add_option("-a", dest="address", 
          help="Local address to bind/connect to", default="localhost")
    return parser




""" Disktools

Methods to manipulate the samplers filesystem
"""

__author__ =  'Walco van Loon'
__version__=  "$Rev$"

import aksy.devices.akai.sysex

class Disktools:
     def __init__(self, connector):
          self.connector = connector
          self.update_disklist_cmd = aksy.devices.akai.sysex.Command('\x5e', '\x10\x01', 'disktools', 'update_disklist', (), None,
              userref_type=aksy.devices.akai.sysex_types.S56K_USERREF)
          self.select_disk_cmd = aksy.devices.akai.sysex.Command('\x5e', '\x10\x02', 'disktools', 'select_disk', (aksy.devices.akai.sysex_types.WORD,), None,
              userref_type=aksy.devices.akai.sysex_types.S56K_USERREF)
          self.test_disk_cmd = aksy.devices.akai.sysex.Command('\x5f', '\x20\x03', 'disktools', 'test_disk', (aksy.devices.akai.sysex_types.WORD,), None)
          self.get_no_disks_cmd = aksy.devices.akai.sysex.Command('\x5f', '\x20\x04', 'disktools', 'get_no_disks', (), None)
          self.get_disklist_cmd = aksy.devices.akai.sysex.Command('\x5e', '\x10\x05', 'disktools', 'get_disklist', (),
              (aksy.devices.akai.sysex_types.DISKLIST,), userref_type=aksy.devices.akai.sysex_types.S56K_USERREF)
          self.get_curr_path_cmd = aksy.devices.akai.sysex.Command('\x5f', '\x20\x09', 'disktools', 'get_curr_path', (), None)
          self.eject_disk_cmd = aksy.devices.akai.sysex.Command('\x5f', '\x20\x0D', 'disktools', 'eject_disk', (aksy.devices.akai.sysex_types.WORD,), None)
          self.get_no_folders_cmd = aksy.devices.akai.sysex.Command('\x5f', '\x20\x10', 'disktools', 'get_no_folders', (), None)
          self.get_folder_names_cmd = aksy.devices.akai.sysex.Command('\x5e', '\x10\x12', 'disktools', 'get_folder_names', (),
              (aksy.devices.akai.sysex_types.STRINGARRAY,), userref_type=aksy.devices.akai.sysex_types.S56K_USERREF)
          self.open_folder_cmd = aksy.devices.akai.sysex.Command('\x5f', '\x20\x13', 'disktools', 'open_folder', (aksy.devices.akai.sysex_types.STRING,), None)
          self.load_folder_cmd = aksy.devices.akai.sysex.Command('\x5e', '\x10\x15', 'disktools', 'load_folder', (aksy.devices.akai.sysex_types.STRING,),
              None, userref_type=aksy.devices.akai.sysex_types.S56K_USERREF)
          self.create_folder_cmd = aksy.devices.akai.sysex.Command('\x5f', '\x20\x16', 'disktools', 'create_folder', (aksy.devices.akai.sysex_types.STRING,), None)
          self.del_folder_cmd = aksy.devices.akai.sysex.Command('\x5f', '\x20\x17', 'disktools', 'delete_folder', (aksy.devices.akai.sysex_types.STRING,), None)
          self.rename_folder_cmd = aksy.devices.akai.sysex.Command('\x5f', '\x20\x18', 'disktools', 'rename_folder', (aksy.devices.akai.sysex_types.STRING, aksy.devices.akai.sysex_types.STRING), None)
          self.get_no_files_cmd = aksy.devices.akai.sysex.Command('\x5e', '\x10\x20', 'disktools', 'get_no_files', (),
              (aksy.devices.akai.sysex_types.WORD,), userref_type=aksy.devices.akai.sysex_types.S56K_USERREF)
          self.get_filenames_cmd = aksy.devices.akai.sysex.Command('\x5e', '\x10\x22', 'disktools', 'get_filenames', (),
              (aksy.devices.akai.sysex_types.NAMESIZEARRAY,), userref_type=aksy.devices.akai.sysex_types.S56K_USERREF)
          self.rename_file_cmd = aksy.devices.akai.sysex.Command('\x5e', '\x10\x28', 'disktools', 'rename_file', (aksy.devices.akai.sysex_types.STRING, aksy.devices.akai.sysex_types.STRING), None, 
              userref_type=aksy.devices.akai.sysex_types.S56K_USERREF)
          self.delete_cmd = aksy.devices.akai.sysex.Command('\x5e', '\x10\x29', 'disktools', 'delete', (aksy.devices.akai.sysex_types.STRING,), None,
              userref_type=aksy.devices.akai.sysex_types.S56K_USERREF)
          self.load_file_cmd = aksy.devices.akai.sysex.Command('\x5e', '\x10\x2A', 'disktools', 'load_file', (aksy.devices.akai.sysex_types.STRING,), None,
              userref_type=aksy.devices.akai.sysex_types.S56K_USERREF)
          self.load_file_and_deps_cmd = aksy.devices.akai.sysex.Command('\x5f', '\x20\x2B', 'disktools', 'load_file_and_deps', (aksy.devices.akai.sysex_types.STRING,), None)
          self.save_cmd = aksy.devices.akai.sysex.Command('\x5e', '\x10\x2C', 'disktools', 'save',
              (aksy.devices.akai.sysex_types.DWORD, aksy.devices.akai.sysex_types.FILETYPE, aksy.devices.akai.sysex_types.BOOL,
               aksy.devices.akai.sysex_types.BOOL), None,
               userref_type=aksy.devices.akai.sysex_types.S56K_USERREF)
          self.save_all_cmd = aksy.devices.akai.sysex.Command('\x5f', '\x20\x2D', 'disktools', 'save_all', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BOOL, aksy.devices.akai.sysex_types.BOOL), None,
              userref_type=aksy.devices.akai.sysex_types.USERREF)

     def update_disklist(self):
          """Update the list of disks connected
          """
          return self.connector.execute(self.update_disklist_cmd, ())

     def select_disk(self, arg0):
          """Select Disk <Data1> = Disk Handle
          """
          return self.connector.execute(self.select_disk_cmd, (arg0, ))

     def test_disk(self, arg0):
          """Test if the disk is valid <Data1> = Disk Handle
          """
          return self.connector.execute(self.test_disk_cmd, (arg0, ))

     def get_no_disks(self):
          """Get the number of disks connected

          Returns:
               aksy.devices.akai.sysex_types.BYTE
          """
          return self.connector.execute(self.get_no_disks_cmd, ())

     def get_disklist(self):
          """Get list of all connected disks

          Returns:
              a list of aksy.devices.akai.sysex_types.DiskInfo objects
          """
          return self.connector.execute(self.get_disklist_cmd, ())

     def get_curr_path(self):
          """Get current path of current disk

          Returns:
               aksy.devices.akai.sysex_types.STRING
          """
          return self.connector.execute(self.get_curr_path_cmd, ())

     def eject_disk(self, arg0):
          """Eject Disk <Data1> = Disk Handle
          """
          return self.connector.execute(self.eject_disk_cmd, (arg0, ))

     def get_no_folders(self):
          """Get number of sub-folders in the current folder.

          Returns:
               aksy.devices.akai.sysex_types.WORD
          """
          return self.connector.execute(self.get_no_folders_cmd, ())

     def get_folder_names(self):
          """Get the names of all of the sub-folders in the current folder.

          Returns:
               aksy.devices.akai.sysex_types.STRINGARRAY
          """
          return self.connector.execute(self.get_folder_names_cmd, ())

     def open_folder(self, arg0):
          """Open Folder. This sets the current folder to be the requested one. (If <Data1> = 0, the root folder will be selected.)
          """
          return self.connector.execute(self.open_folder_cmd, (arg0, ))

     def load_folder(self, arg0):
          """Load Folder: the selected folder, and all its contents (including folders)
          """
          return self.connector.execute(self.load_folder_cmd, (arg0, ))

     def create_folder(self, arg0):
          """Create Folder: Creates a folder in the currently selected folder.
          """
          return self.connector.execute(self.create_folder_cmd, (arg0, ))

     def delete_folder(self, arg0):
          """Delete Folder.
          """
          return self.connector.execute(self.del_folder_cmd, (arg0, ))

     def rename_folder(self, arg0, arg1):
          """Rename Folder: <Data1> = name of folder to rename
          """
          return self.connector.execute(self.rename_folder_cmd, (arg0, arg1, ))

     def get_no_files(self):
          """Get number of files in the current folder.

          Returns:
               aksy.devices.akai.sysex_types.WORD
          """
          return self.connector.execute(self.get_no_files_cmd, ())

     def get_filenames(self):
          """Get the names of all of the files in the current folder.

          Returns:
               aksy.devices.akai.sysex_types.STRINGARRAY
          """
          return self.connector.execute(self.get_filenames_cmd, ())

     def rename_file(self, arg0, arg1):
          """Rename File
          """
          return self.connector.execute(self.rename_file_cmd, (arg0, arg1, ))

     def delete(self, arg0):
          """Delete file or folder (including folder content!)
          """
          return self.connector.execute(self.delete_cmd, (arg0, ))

     def load_file(self, arg0):
          """Load File <Data1> = name of file to load.
          """
          return self.connector.execute(self.load_file_cmd, (arg0, ))

     def load_file_and_deps(self, arg0):
          """Load File <Data1> = name of file to load. Will load the dependents as well
          """
          return self.connector.execute(self.load_file_and_deps_cmd, (arg0, ))

     def save(self, arg0, arg1, arg2, arg3):
          """Save Memory Item to Disk
          handle Handle of Memory Item
          type (1=Multi; 2=Program; 3=Sample; 4=SMF)
          overwrite (0=Skip if file exists; 1=Overwrite existing files)
          save_childen = (0=Don't save children; 1=Save Children)
          
          """
          return self.connector.execute(self.save_cmd, (arg0, arg1, arg2, arg3, ))

     def save_all(self, arg0, arg1, arg2):
          """Save All Memory Items to Disk
          Type = (0=All; 1=Multi; 2=Program; 3=Sample; 4=SMF)
          Overwrite (0=Skip if file exists; 1=Overwrite existing files)
          SaveChildren = (0=Don't save children; 1=Save Children)
          """
          return self.connector.execute(self.save_all_cmd, (arg0, arg1, arg2))


import gtk

class AkWidget(gtk.DrawingArea):
    """
    base for our widgets
    """
    def __init__(self, so = None, soattr = None, interval = 10, units = "", value_offset = 0):
        gtk.DrawingArea.__init__(self)

        self.value_offset = value_offset
        self.interval = interval
        self.units = units

        self.connect("expose_event", self.on_expose)
        self.connect("button_press_event", self.on_button_press)
        self.connect("button_release_event", self.on_button_release)
        self.connect("motion_notify_event", self.on_motion_notify_event)

        self.add_events(gtk.gdk.BUTTON_PRESS_MASK |
                        gtk.gdk.BUTTON_RELEASE_MASK |
                        gtk.gdk.POINTER_MOTION_MASK)

        self.is_scaled = False

        self.so = None
        self.soattr = None

        self.init(so, soattr)

    def init(self, so, soattr = None):
        if soattr:
            self.soattr = soattr

        if so:
            self.s = so.s
            self.so = so
            self.so.precache()

            if self.soattr:
                self.value = getattr(self.so, self.soattr)
                
            #self.init_filechooser(self.s)

        self.queue_draw()

    def get_format(self):
        if self.units:
            return '%(#).2f%(units)s' % {"#" : self.value / self.interval, "units" : self.units}
        elif self.interval > 1:
            return '%(#).2f' % {"#" : self.value / self.interval}
        else:
            return str(self.value)

    def init_filechooser(self, s):
        if not getattr(self.s, 'filechooser', None):
            setattr(self.s,'filechooser', UI.FileChooser(s))

    def set_value(self, value):
        # returns True if value actually changed

        before = self.value
        self.value = self.calc_value(value)
        
        if before == self.value:
            return False
        else:
            return True

    def calc_value(self, value):
        # scale and restrict value

        if self.is_scaled:
            scale = self.get_scale()
            newvalue = int(value * scale)
        else:
            newvalue = value

        if newvalue <= self.softmax and newvalue >= self.softmin:
            return newvalue
        elif newvalue > self.softmax:
            return self.softmax
        elif newvalue < self.softmin:
            return self.softmin

    def on_button_press(self, widget, event):
        """
        Example
        for x,y in points:
            hb = HitBox(x-5,y-5,10,10)

            if hb.point_in(event.x,event.y): 
                self.dragging = points.index([x,y])
        """

    def on_expose(self, widget, event):
        """
        self.context = widget.window.cairo_create()
        self.context.set_source_rgb(1, 1, 1)
        self.context.fill_preserve()
        self.context.set_source_rgb(0, 0, 0)
        self.context.stroke()
        """



import samplerobject, zone, envelope

keygroup_cache = {}

def get_keygroup_cached(p, index):
    key = (p.handle, index)

    if not key in keygroup_cache.keys():
        keygroup_cache[key] = Keygroup(p, index) 

    keygroup = keygroup_cache[key]

    return keygroup

class Keygroup(samplerobject.SamplerObject):
    def __init__(self, program, index, precache_on_init = True):
        samplerobject.SamplerObject.__init__(self, program.s, None, "keygrouptools")
        self.current_mod_source_index = 11
        self.attrs = ["low_note", "high_note", "mute_group", "fx_override", "fx_send_level", "zone_xfade", "zone_xfade_type", "polyphony", "tune", "level", "play_trigger", "play_trigger_velocity", "filter", "filter_cutoff", "filter_resonance", "filter_attenuation"]
        self.attrs_minimal = ["low_note", "high_note"]
        self.abbr = {'polyphony' : 'poly', 'filter_cutoff':'cutoff', 'filter_resonance':'res', 'MOD_12_14':'filtenv', 'MOD_12_15':'filtenv', 'MOD_6_1':'velo', 'MOD_11_1':'ampenv', 'MOD_13_3':'auxenv',}
        self.p = program
        self.index = index
        self.keygroup_index = index
        self.s = program.s
        
        if getattr(self.s, "samples", None):
            self.samples = self.s.samples
        else:
            self.samples = None
            
        kgt = self.s.keygrouptools

        self.filter_attributes = ['filter', 'filter_cutoff', 'filter_resonance']

        self.amp_envelope = envelope.Envelope(self, 0)
        self.filter_envelope = envelope.Envelope(self, 1)
        self.aux_envelope = envelope.Envelope(self, 2)

        self.zones = [zone.Zone(self, 1), zone.Zone(self, 2), zone.Zone(self, 3), zone.Zone(self, 4)]
        self.mod_matrix = []
        self.mod_matrix_dict = {}
       
        if precache_on_init:
            self.precache()

    def load_matrix(self):
        (self.mod_matrix, self.mod_matrix_dict) = self.p.get_matrix(self.index + 1)

    def get_handle(self):
        # get handle by name or whatever, override in derived classes
        return self.p.handle

    def set_current(self):
        self.s.keygrouptools.set_curr_keygroup(self.index)

    def on_mod_source_changed(self, widget):
        self.current_mod_source_index = widget.get_active()

    """
    Need to phase this override out and merge it into samplerobject.set
    """
    def set(self, attrname, attrval):
        if attrname.startswith("MOD_"):
            pin = self.get_pin_by_name(attrname, True)
            self.attrscache[attrname] = pin.set_value(attrval)
            return
        
        kgt = self.s.keygrouptools
        kgt.set_curr_keygroup(self.keygroup_index)
        func = getattr(kgt, "set_" + attrname, None)
        if func:
            if attrname in self.filter_attributes:
                func(0, attrval)
            else:
                if isinstance(attrval, float):
                    attrval = int(attrval)
                func(attrval)
        else:
            print "no method set_" + attrname

        if self.set_callback:
            self.set_callback(attrname, attrval)
            
        self.attrscache[attrname] = attrval
        
    def set_pin_value(self, source, value):
        if len(self.mod_matrix) > 0:
            return self.mod_matrix[source]
        else:
            return None

    def get_next_empty_pin(self):
        for pin in self.mod_matrix:
            if pin.dest == 0:
                return pin
        return None

    def get_pin_by_name(self, attrname, create_new = False):
        if attrname.startswith("MOD_"):
            (blah, source, dest) = attrname.split("_")
            return self.get_pin_by_source_and_dest(int(source), int(dest), create_new)
        
    def get_pin_by_source_and_dest(self, source, dest, create_new = False):
        pin = None

        source_dest_index = (source, dest)

        if source_dest_index in self.mod_matrix_dict:
            pin = self.mod_matrix[self.mod_matrix_dict[source_dest_index].pin_index]
        elif create_new:
            # didnt find a pin!
            pin = self.get_next_empty_pin()
            pin.source = source
            pin.dest = dest
            self.mod_matrix_dict[(source, dest)] = pin

        return pin



import os, struct
from StringIO import StringIO
from aksy.devices.akai.filemodel import Zone, Program, Chunk, Keygroup

def parse_byte(chunk, offset):
    return 1, struct.unpack('<b', chunk[offset])

def parse_string(chunk, offset, length=None):
    if length is None:
        length = struct.unpack('B', chunk[offset:offset+1])[0]
        offset += 1
        
    return length, struct.unpack('%is' %length, chunk[offset: offset+length])[0]
        
class ChunkParser:
    def parse(self, fh, offset):
        bytes_read = fh.read(4)
        assert len(bytes_read) == 4
        name = parse_string(bytes_read, 0, 4)[1]
        bytes_read = fh.read(4)
        assert len(bytes_read)
        length = struct.unpack('<l', bytes_read)[0]
        bytes = fh.read(length)
        return Chunk(name, bytes, length+8)

    def parse_chunks(self, fh, offset, end):
        chunks = []
        while offset < end:
            chunk = self.parse(fh, offset)
            offset += chunk.length
            chunks.append(chunk)
        return chunks


def get_file_length(filename):
        return os.stat(filename).st_size
    
class ProgramParser:
    def __init__(self):
        self.chunkParser = ChunkParser()
        self.keygroupParser = KeygroupParser()
        self.chunks = []
        
    def parse_chunks(self, filelength):
        # skip RIFF header
        offset = 0x0c
        self.fh.seek(offset)
        return self.chunkParser.parse_chunks(self.fh, offset, filelength)

    def parse_keygroups(self):
        keygroups = []
        for chunk in self.chunks:
            if chunk.name == "kgrp":
                keygroups.append(self.keygroupParser.parse(chunk))
        return keygroups
            
    def parse(self, filename):
        self.fh = file(filename, 'rb')
        self.chunks = self.parse_chunks(get_file_length(filename))
        prg = Program(self.parse_keygroups())
        return prg

class KeygroupParser:
    def __init__(self):
        self.chunkParser = ChunkParser()
        self.zoneParser = ZoneParser()
        self.chunks = []
        
    def parse_zones(self, chunks):
        zones = []
        for chunk in chunks:
            if chunk.name == "zone":
                zones.append(self.zoneParser.parse(chunk))
        return zones
    
    def parse(self, chunk):
        chunks = self.chunkParser.parse_chunks(StringIO(chunk.bytes), 0, chunk.get_content_length())
        kg = Keygroup(self.parse_zones(chunks))
        return kg
    
class ZoneParser:
    def parse(self, chunk):
        zone = Zone()
        zone.samplename = parse_string(chunk.bytes, 1)[1]
        return zone


import samplerobject

from aksy import fileutils

class Recording(samplerobject.SamplerObject):
    def __init__(self, s):
        samplerobject.SamplerObject.__init__(self, s, None, "recordingtools")
        self.attrs = ["status", "progress", "max_rec_time", "input", "mode",
"monitor", "rec_time", "pitch", "threshold", "trigger_src", "bit_depth",
"prerec_time", "name", "name_seed", "autorec_mode", "autonormalize"]
        if not fileutils.is_valid_name(self.s.recordingtools.get_name()):
            self.set("name", "Sample %i" % (self.s.sampletools.get_no_items() + 1))
        self.update()

    def set(self, attrname, attrval):
        """ Overrides handle based lookups in base class
        """
        tools = self.gettools()
        getattr(tools, "set_" + attrname)(attrval)





import gtk

from aksui.ak import multifx
from aksui.utils import modelutils

class MultiFXChannelEditor(gtk.VBox):
    def __init__(self,mMultiFXChannelEditor):
        gtk.VBox.__init__(self)
        self.mMultiFXChannelEditor = mMultiFXChannelEditor
        self.channel = mMultiFXChannelEditor

        self.moduleWidgets = { }

        self.topHBox = gtk.HBox(True)

        self.inputComboBox = modelutils.magicCombo(multifx.inputmapmodel,self.channel.input,self.on_input_changed,2)
        self.topHBox.pack_start(self.inputComboBox)

        self.outputComboBox = modelutils.magicCombo(multifx.outputmapmodel,self.channel.output,self.on_output_changed,3)
        self.topHBox.pack_start(self.outputComboBox)

        self.pack_start(self.topHBox)

        for i in range(self.mMultiFXChannelEditor.max_modules):
            self.moduleWidgets[i] = MultiFXModuleEditor(mMultiFXChannelEditor.modules[i])
            self.pack_start(self.moduleWidgets[i], expand=False, fill=False)

    def on_input_changed(self, widget):
        self.channel.updateInput(widget.get_active())

    def on_output_changed(self, widget):
        self.channel.updateOutput(widget.get_active())

class MultiFXModuleEditor(gtk.VBox):
    def __init__(self,module):
        gtk.VBox.__init__(self)
        self.module = module

        self.parameterWidgets = { }

        self.topHBox = gtk.HBox()

        self.effectsComboBox = modelutils.magicCombo(multifx.effectmodel,self.module.effect_index,self.on_changed,4)
        self.topHBox.pack_start(self.effectsComboBox)

        self.enableButton = gtk.ToggleButton('Enable')

        if self.module.enabled:
            self.enableButton.set_active(1)
        else:
            self.enableButton.set_active(0)

        self.enableButton.connect('toggled', self.on_toggle)

        self.topHBox.pack_start(self.enableButton, expand=False, fill=False)

        self.pack_start(self.topHBox)
        self.updateWidgets()

    def on_toggle(self, widget):
        if(widget.get_active()):
            self.module.enable()
        else:
            self.module.disable()

    def on_changed(self, widget):
        active_index = widget.get_active()
        self.module.setEffectById(active_index)
        self.updateWidgets()

    def updateWidgets(self):
        module = self.module

        for i in self.parameterWidgets:
            self.parameterWidgets[i].destroy()

        self.parameterWidgets = { }

        for i in module.parameters:
            w = MultiFXParameter(module.parameters[i])
            self.parameterWidgets[i] = w

        for i in self.parameterWidgets:
            w = self.parameterWidgets[i]
            self.pack_start(w, expand=False, fill=False)

        self.show_all()

class MultiFXParameter(gtk.HBox):
    def __init__(self,param):
        gtk.HBox.__init__(self, True)
        self.param = param

        self.lblFormat = gtk.Label(param.format_value())

        paramap = {'name':gtk.Label, 'value':gtk.HScale}

        for n in paramap:
            t = paramap[n]

            if t == gtk.Label:
                arg0 = getattr(param,n)
            elif t == gtk.HScale:
                arg0 = gtk.Adjustment(int(param.value),int(param.minimum),int(param.maximum)+1,0.01,1,1)
                arg0.connect('value_changed', param.on_value_changed, self.lblFormat)

            setattr(self,'_param'+n, t(arg0))
            o = getattr(self,'_param'+n)

            if t == gtk.HScale:
                o.set_digits(0)
                o.set_value_pos(gtk.POS_BOTTOM)
                o.set_draw_value(False)

            self.pack_start(o)

        self.pack_start(self.lblFormat)

class MultiFXLabel(gtk.Label):
    def __init__(self,text):
        gtk.Label.__init__(self,text)
        self.justify = gtk.JUSTIFY_LEFT

class MultiFXEditor(gtk.VBox):
    def __init__(self,s):
        gtk.VBox.__init__(self)
        self.mfx = multifx.MultiFX(s)

        self.channelWidgets = { }
      
        if self.mfx.channels:
            for i in self.mfx.channels:
                self.channelWidgets[i] = MultiFXChannelEditor(self.mfx.channels[i])
                self.pack_start(self.channelWidgets[i], expand=False, fill=False)
        else:
            print "nop"




# package aksy
import logging, logging.config, logging.handlers,  time, os.path, sys

from aksy.config import get_config

# initialize logging
logger = logging.getLogger('aksy')
level = getattr(logging, get_config().get('logging', 'level'))
logger.setLevel(level)

logdir = os.path.expanduser('~/.aksy/logs')
if not os.path.exists(logdir):
    os.makedirs(logdir)

script_name = os.path.basename(sys.argv[0])
handler = logging.FileHandler(os.path.expanduser('~/.aksy/logs/%s_%s.log' % (script_name, time.strftime("%Y%m%d"))))
formatter = logging.Formatter(fmt="%(asctime)s %(levelname)s %(name)s:%(lineno)d %(message)s")
handler.setFormatter(formatter)

logger.addHandler(handler)




""" Python equivalent of akai section multitools

Methods to manipulate in-memory multis
"""

__author__ =  'Walco van Loon'
__version__ =  '0.2'

from aksy.devices.akai.sysex import Command

import aksy.devices.akai.sysex_types

class Multitools:
    def __init__(self, s56k):
        self.sampler = s56k
        self.rename_cmd = Command('^', '\x0c\x30', 'multitools', 'rename', (aksy.devices.akai.sysex_types.STRING,), (), aksy.devices.akai.sysex_types.S56K_USERREF)
        self.get_no_items_cmd = Command('^', '\x0c\x40', 'multitools', 'get_no_items', (), (aksy.devices.akai.sysex_types.CWORD,), aksy.devices.akai.sysex_types.S56K_USERREF)
        self.get_curr_handle_cmd = Command('^', '\x0c\x42', 'multitools', 'get_curr_handle', (), (aksy.devices.akai.sysex_types.CWORD,), aksy.devices.akai.sysex_types.S56K_USERREF)
        self.get_curr_name_cmd = Command('^', '\x0c\x43', 'multitools', 'get_curr_name', (), (aksy.devices.akai.sysex_types.STRINGARRAY,), aksy.devices.akai.sysex_types.S56K_USERREF)
        self.get_names_cmd = Command('^', '\x0c\x51', 'multitools', 'get_names', (), (aksy.devices.akai.sysex_types.STRINGARRAY,), aksy.devices.akai.sysex_types.S56K_USERREF)
        self.set_curr_by_handle_cmd = Command('^', '\x0c\x06', 'multitools', 'set_curr_by_handle', (aksy.devices.akai.sysex_types.CWORD,), (), aksy.devices.akai.sysex_types.S56K_USERREF)
        self.set_curr_by_name_cmd = Command('^', '\x0c\x05', 'multitools', 'set_curr_by_name', (aksy.devices.akai.sysex_types.STRING,), (), aksy.devices.akai.sysex_types.S56K_USERREF)
        self.delete_all_cmd = Command('^', '\x0c\x07', 'multitools', 'delete_all', (), (), aksy.devices.akai.sysex_types.S56K_USERREF)
        self.delete_curr_cmd = Command('^', '\x0c\x08', 'multitools', 'delete_curr', (), (), aksy.devices.akai.sysex_types.S56K_USERREF)

    def rename(self):
        """Rename multi
        """
        return self.sampler.execute(self.rename_cmd, ())

    def get_no_items(self):
        """Get Number of Multis in Memory

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_no_items_cmd, ())

    def get_curr_handle(self):
        """Current Multi Index (i.e., its position in memory)

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_curr_handle_cmd, ())

    def get_curr_name(self):
        """Get Current Multi Name

        Returns:
            STRING
        """
        return self.sampler.execute(self.get_curr_name_cmd, ())

    def get_names(self):
        """Get the Names of All Multis in memory

        Returns:
            STRINGARRAY
        """
        return self.sampler.execute(self.get_names_cmd, ())

    def set_curr_by_handle(self, arg0):
        """Select by handle
        """
        return self.sampler.execute(self.set_curr_by_handle_cmd, (arg0, ))

    def set_curr_by_name(self, arg0):
        """Select by name
        """
        return self.sampler.execute(self.set_curr_by_name_cmd, (arg0, ))

    def delete_all(self):
        """Delete ALL items from memory
        """
        return self.sampler.execute(self.delete_all_cmd, ())

    def delete_curr(self):
        """Delete current item from memory
        """
        return self.sampler.execute(self.delete_curr_cmd, ())

    def get_handles_names(self):
        """Get list of multi handles and names

        Returns:
            HANDLENAMEARRAY
        """
        handle_names = []
        names = self.get_names()
        for i in range(len(names)):
            handle_names.extend([i, names[i]])
        return handle_names





""" Python equivalent of akai section songtools

Song
"""

__author__ =  'Walco van Loon'
__version__ =  '0.2'

from aksy.devices.akai.sysex import Command

import aksy.devices.akai.sysex_types

class Songtools:
    def __init__(self, z48):
        self.sampler = z48
        self.get_no_items_cmd = Command('_', '\x28\x01', 'songtools', 'get_no_items', (), None)
        self.get_handles_cmd = Command('_', '\x28\x02\x00', 'songtools', 'get_handles', (), None)
        self.get_names_cmd = Command('_', '\x28\x02\x01', 'songtools', 'get_names', (), None)
        self.get_handles_names_cmd = Command('_', '\x28\x02\x02', 'songtools', 'get_handles_names', (), None)
        self.get_handles_modified_cmd = Command('_', '\x28\x02\x03', 'songtools', 'get_handles_modified', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_curr_by_handle_cmd = Command('_', '\x28\x03', 'songtools', 'set_curr_by_handle', (aksy.devices.akai.sysex_types.DWORD,), None)
        self.set_curr_by_name_cmd = Command('_', '\x28\x04', 'songtools', 'set_curr_by_name', (aksy.devices.akai.sysex_types.STRING,), None)
        self.get_curr_handle_cmd = Command('_', '\x28\x05', 'songtools', 'get_curr_handle', (), None)
        self.get_curr_name_cmd = Command('_', '\x28\x06', 'songtools', 'get_curr_name', (), None)
        self.get_name_by_handle_cmd = Command('_', '\x28\x07', 'songtools', 'get_name_by_handle', (aksy.devices.akai.sysex_types.DWORD,), None)
        self.get_handle_by_name_cmd = Command('_', '\x28\x08', 'songtools', 'get_handle_by_name', (aksy.devices.akai.sysex_types.STRING,), None)
        self.delete_all_cmd = Command('_', '\x28\x09', 'songtools', 'delete_all', (), None)
        self.delete_curr_cmd = Command('_', '\x28\x0A', 'songtools', 'delete_curr', (), None)
        self.delete_by_handle_cmd = Command('_', '\x28\x0B', 'songtools', 'delete_by_handle', (aksy.devices.akai.sysex_types.DWORD,), None)
        self.rename_curr_cmd = Command('_', '\x28\x0C', 'songtools', 'rename_curr', (aksy.devices.akai.sysex_types.STRING,), None)
        self.rename_by_handle_cmd = Command('_', '\x28\x0D', 'songtools', 'rename_by_handle', (aksy.devices.akai.sysex_types.DWORD, aksy.devices.akai.sysex_types.STRING), None)
        self.set_tag_bit_cmd = Command('_', '\x28\x0E', 'songtools', 'set_tag_bit', (aksy.devices.akai.sysex_types.BYTE, aksy.devices.akai.sysex_types.BYTE), None)
        self.get_tag_bitmap_cmd = Command('_', '\x28\x0F', 'songtools', 'get_tag_bitmap', (), None)
        self.get_curr_modified_cmd = Command('_', '\x28\x10', 'songtools', 'get_curr_modified', (), None)
        self.get_modified_cmd = Command('_', '\x28\x11', 'songtools', 'get_modified', (), None)
        self.delete_tagged_cmd = Command('_', '\x28\x18', 'songtools', 'delete_tagged', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.play_song_cmd = Command('_', '\x28\x40', 'songtools', 'play_song', (), None)
        self.pause_song_cmd = Command('_', '\x28\x41', 'songtools', 'pause_song', (), None)
        self.stop_song_cmd = Command('_', '\x28\x42', 'songtools', 'stop_song', (), None)
        self.set_group_id_cmd = Command('_', '\x28\x01', 'songtools', 'set_group_id', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_from_bar_cmd = Command('_', '\x28\x10', 'songtools', 'set_from_bar', (aksy.devices.akai.sysex_types.WORD,), None)
        self.set_to_bar_cmd = Command('_', '\x28\x10', 'songtools', 'set_to_bar', (aksy.devices.akai.sysex_types.WORD,), None)
        self.set_tempo_mode_cmd = Command('_', '\x28\x12', 'songtools', 'set_tempo_mode', (aksy.devices.akai.sysex_types.BYTE,), None)
        self.set_manual_tempo_cmd = Command('_', '\x28\x13', 'songtools', 'set_manual_tempo', (aksy.devices.akai.sysex_types.WORD,), None)
        self.set_midi_output_cmd = Command('_', '\x28\x18', 'songtools', 'set_midi_output', (aksy.devices.akai.sysex_types.BOOL,), None)
        self.get_group_id_cmd = Command('_', '\x28\x01', 'songtools', 'get_group_id', (), None)
        self.get_from_bar_cmd = Command('_', '\x28\x10', 'songtools', 'get_from_bar', (), None)
        self.get_to_bar_cmd = Command('_', '\x28\x11', 'songtools', 'get_to_bar', (), None)
        self.get_tempo_mode_cmd = Command('_', '\x28\x12', 'songtools', 'get_tempo_mode', (), None)
        self.get_manual_tempo_cmd = Command('_', '\x28\x13', 'songtools', 'get_manual_tempo', (), None)
        self.get_midi_output_port_cmd = Command('_', '\x28\x18', 'songtools', 'get_midi_output_port', (), None)
        self.get_time_signature_beat_cmd = Command('_', '\x28\x20', 'songtools', 'get_time_signature_beat', (), None)
        self.get_time_sig_beat_no_cmd = Command('_', '\x28\x21', 'songtools', 'get_time_sig_beat_no', (), None)
        self.get_curr_beat_cmd = Command('_', '\x28\x22', 'songtools', 'get_curr_beat', (), None)
        self.get_curr_bar_cmd = Command('_', '\x28\x23', 'songtools', 'get_curr_bar', (), None)
        self.get_curr_tempo_cmd = Command('_', '\x28\x24', 'songtools', 'get_curr_tempo', (), None)

    def get_no_items(self):
        """Get number of items in memory

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_no_items_cmd, ())

    def get_handles(self):
        """Get Sample handles
        """
        return self.sampler.execute(self.get_handles_cmd, ())

    def get_names(self):
        """Get sample names

        Returns:
            STRINGARRAY
        """
        return self.sampler.execute(self.get_names_cmd, ())

    def get_handles_names(self):
        """Get list of sample handles and names

        Returns:
            HANDLENAMEARRAY
        """
        return self.sampler.execute(self.get_handles_names_cmd, ())

    def get_handles_modified(self, arg0):
        """Get a list of modified/tagged samples
        """
        return self.sampler.execute(self.get_handles_modified_cmd, (arg0, ))

    def set_curr_by_handle(self, arg0):
        """Select current item by handle
        """
        return self.sampler.execute(self.set_curr_by_handle_cmd, (arg0, ))

    def set_curr_by_name(self, arg0):
        """Select current item by name
        """
        return self.sampler.execute(self.set_curr_by_name_cmd, (arg0, ))

    def get_curr_handle(self):
        """Get handle of current item

        Returns:
            DWORD
        """
        return self.sampler.execute(self.get_curr_handle_cmd, ())

    def get_curr_name(self):
        """Get name of current item

        Returns:
            STRING
        """
        return self.sampler.execute(self.get_curr_name_cmd, ())

    def get_name_by_handle(self, arg0):
        """Get item name from handle

        Returns:
            STRING
        """
        return self.sampler.execute(self.get_name_by_handle_cmd, (arg0, ))

    def get_handle_by_name(self, arg0):
        """Get item handle from name

        Returns:
            DWORD
        """
        return self.sampler.execute(self.get_handle_by_name_cmd, (arg0, ))

    def delete_all(self):
        """Delete ALL items from memory
        """
        return self.sampler.execute(self.delete_all_cmd, ())

    def delete_curr(self):
        """Delete current item from memory
        """
        return self.sampler.execute(self.delete_curr_cmd, ())

    def delete_by_handle(self, arg0):
        """Delete item represented by handle <Data1>
        """
        return self.sampler.execute(self.delete_by_handle_cmd, (arg0, ))

    def rename_curr(self, arg0):
        """Rename current item
        """
        return self.sampler.execute(self.rename_curr_cmd, (arg0, ))

    def rename_by_handle(self, arg0, arg1):
        """Rename item represented by handle <Data1>
        """
        return self.sampler.execute(self.rename_by_handle_cmd, (arg0, arg1, ))

    def set_tag_bit(self, arg0, arg1):
        """Set Tag Bit <Data1> = bit to set, <Data2> = (0=OFF, 1=ON) BYTE(0, 1) <Data3> = (0=CURRENT, 1=ALL)
        """
        return self.sampler.execute(self.set_tag_bit_cmd, (arg0, arg1, ))

    def get_tag_bitmap(self):
        """Get Tag Bitmap
        """
        return self.sampler.execute(self.get_tag_bitmap_cmd, ())

    def get_curr_modified(self):
        """Get name of current item with modified/tagged info

        Returns:
            STRINGARRAY
        """
        return self.sampler.execute(self.get_curr_modified_cmd, ())

    def get_modified(self):
        """Get modified state of current item.

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_modified_cmd, ())

    def delete_tagged(self, arg0):
        """Delete tagged items <Data1> = tag bit
        """
        return self.sampler.execute(self.delete_tagged_cmd, (arg0, ))

    def play_song(self):
        """Play Song
        """
        return self.sampler.execute(self.play_song_cmd, ())

    def pause_song(self):
        """Pause Song
        """
        return self.sampler.execute(self.pause_song_cmd, ())

    def stop_song(self):
        """Stop Song
        """
        return self.sampler.execute(self.stop_song_cmd, ())

    def set_group_id(self, arg0):
        """Set Group ID
        """
        return self.sampler.execute(self.set_group_id_cmd, (arg0, ))

    def set_from_bar(self, arg0):
        """Set From Bar
        """
        return self.sampler.execute(self.set_from_bar_cmd, (arg0, ))

    def set_to_bar(self, arg0):
        """Set To Bar
        """
        return self.sampler.execute(self.set_to_bar_cmd, (arg0, ))

    def set_tempo_mode(self, arg0):
        """Set Tempo Mode <Data1> = (0=FILE, 1=MANUAL, 2=MULTI)
        """
        return self.sampler.execute(self.set_tempo_mode_cmd, (arg0, ))

    def set_manual_tempo(self, arg0):
        """Set Manual Tempo <Data1> = (tempo*10)bpm
        """
        return self.sampler.execute(self.set_manual_tempo_cmd, (arg0, ))

    def set_midi_output(self, arg0):
        """Set MIDI output port <Data1> = (0=NONE, 1=MIDI A, 2=MIDI B)
        """
        return self.sampler.execute(self.set_midi_output_cmd, (arg0, ))

    def get_group_id(self):
        """Get Group ID

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_group_id_cmd, ())

    def get_from_bar(self):
        """Get From Bar

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_from_bar_cmd, ())

    def get_to_bar(self):
        """Get To Bar

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_to_bar_cmd, ())

    def get_tempo_mode(self):
        """Get Tempo Mode <Reply> = (0=FILE, 1=MANUAL, 2=MULTI)

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_tempo_mode_cmd, ())

    def get_manual_tempo(self):
        """Get Manual Tempo

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_manual_tempo_cmd, ())

    def get_midi_output_port(self):
        """Get MIDI output port <Reply> = (0=NONE, 1=MIDI A, 2=MIDI B

        Returns:
            BOOL
        """
        return self.sampler.execute(self.get_midi_output_port_cmd, ())

    def get_time_signature_beat(self):
        """Get (Time Signature) Beat Value

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_time_signature_beat_cmd, ())

    def get_time_sig_beat_no(self):
        """Get (Time Signature) Beats-per-Bar

        Returns:
            BYTE
        """
        return self.sampler.execute(self.get_time_sig_beat_no_cmd, ())

    def get_curr_beat(self):
        """Get Current Beat

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_curr_beat_cmd, ())

    def get_curr_bar(self):
        """Get Current Bar

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_curr_bar_cmd, ())

    def get_curr_tempo(self):
        """Get Current Tempo <Reply> = (tempo*10)bpm

        Returns:
            WORD
        """
        return self.sampler.execute(self.get_curr_tempo_cmd, ())

